# NOTE: The configuration file puma.config depends on the exact version of gcc
# that generated it.  However, the library itself does not.  To avoid the
# necessity of recompiling libpuma for every gcc update, we instead generate
# the configuration file on the fly as the installed gcc is updated.  See the
# trigger script below.
#
# Since the trigger script requires the presence of the ag++ binary, the binary
# and the library thus depend on each other, so we collapse them into a single
# package.  This means multilib is currently impossible, but it was impossible
# anyway, since we would have needed a multilib ag++ binary to generate the
# multilib config.

Name:           libpuma
Version:        1.2
Release:        3%{?dist}
Summary:        Library for parsing and manipulating C/C++ source code

License:        GPLv2+
URL:            http://aspectc.org/
Source0:        http://aspectc.org/releases/%{version}/ac-woven-%{version}.tar.gz
# Man pages from Debian
Source1:        ac++.1
Source2:        ag++.1
# http://www.aspectc.org/bugzilla/show_bug.cgi?id=519
Patch0:         %{name}-doxygen.patch
# Fedora-specific: build a shared library instead of a static library.
Patch1:         %{name}-shared.patch
# Fedora-specific: fix a few tests
Patch2:         %{name}-test.patch
# Fedora-specific: adapt to the newer lexertl in Fedora.
Patch3:         %{name}-lexertl.patch

BuildRequires:  doxygen
BuildRequires:  graphviz
BuildRequires:  lexertl-devel
BuildRequires:  libxml2-devel

Requires:       gcc-c++
Requires:       libstdc++-devel%{?_isa}
Requires:       glibc-devel%{?_isa}

# This can be removed once F-21 reaches EOL.
Obsoletes:      aspectc++ < 1.2-3%{?dist}
Provides:       aspectc++ = %{version}-%{release}

%description
PUMA is a library of C++ classes for parsing and manipulating C/C++
source code.  It provides different levels of analyzing source code,
i.e. lexical, syntactical, and semantical analyses.  Additionally, PUMA
contains its own C preprocessor following the macro language understood
by the GNU C preprocessor (cpp).

The parsers for C and C++ are based on an extensible recursive descent
parser architecture that can be easily adapted for your own needs by
inheriting from one of the parsers.  They are oriented on the ISO/IEC
standards defined in ISO/IEC 9899:1999(E) and ISO/IEC 14882:1998(E) but
also supporting a list of extensions to the C and C++ languages.  The
information about the source code being analyzed is organized in an
attributed syntax tree referring to a separate structure of semantic
information.

For the purpose of manipulating source code PUMA provides a syntax based
mechanism for searching code structures.  Manipulations on the source
code can be collected and validated before executing them in a single
transaction.

%package devel
Summary:        Development files for %{name}
Requires:       %{name}%{?_isa} = %{version}-%{release}
Requires:       lexertl-devel

%description devel
Header files and library links for developing applications that use
AspectC++.

%package doc
Summary:        Documentation for %{name}
BuildArch:      noarch

%description doc
User and developer documentation for %{name}.

%prep
%setup -q -c
%patch0
%patch1
%patch2
%patch3

# Remove prebuilt objects
rm -f aspectc++/Puma/tools/lemon/lemon

# Show compiler flags
sed -i 's/@\$(C/$(C/' aspectc++/Puma/src/compile.mk \
    aspectc++/AspectC++/Makefile aspectc++/AspectC++/ACModel/Makefile \
    aspectc++/Ag++/Makefile

# Fix an install directory
if [ "%{_libdir}" = "%{_prefix}/lib64" ]; then
  sed -i 's,\$(PREFIX)/lib,$(PREFIX)/lib64,' aspectc++/Puma/Makefile
fi

# Use the system lexertl instead of the bundled version
rm -fr aspectc++/Puma/extern/lexertl
cp -a %{_includedir}/lexertl aspectc++/Puma/extern/lexertl

# Pass build flags to the linker, and set an soname
sed "s|-shared|& %{optflags} $RPM_LD_FLAGS -Wl,-h,libPuma.so.0 -Wl,--as-needed|" \
    -i aspectc++/Puma/src/compile.mk

# Change the default location of the config file, for multilib builds
sed -i 's,/etc/puma.config,%{_libdir}/%{name}/puma.config,' \
    aspectc++/Puma/gen-release/step1/src/Config.cc \
    aspectc++/Puma/gen-release/step2/src/Config.cc \
    aspectc++/Puma/src/common/Config.cc

# Fix a test that is broken on 64-bit systems
%if %{__isa_bits} == 64
sed -i 's/unsigned int/unsigned long int/' \
    aspectc++/AspectC++/tests/ExecAdviceNewDelete/feature.ref
%endif

# Do the initial build of the prewoven sources in a separate build dir
cp -a aspectc++ aspectc++.prewoven

# Avoid a name collision in the docs
cp -p aspectc++/README aspectc++/README.aspectc++

%build
# Phase 1: bootstrap by building the tools with the prewoven sources.
pushd aspectc++.prewoven

# Support __int128 if gcc supports it
if gcc -dM -E - <<<'' | grep -Fq __SIZEOF_INT128__; then
  export AC_OPTFLAGS=--gnu-int128
fi
BUILDFLAGS="%{optflags} -fsigned-char -Wno-unused-local-typedefs $RPM_LD_FLAGS"

# Build the bootstrap binaries
make %{?_smp_mflags} -C Puma compile CFLAGS="$BUILDFLAGS -fPIC"
ln -s libPuma.so Puma/lib/linux-release/libPuma.so.0
make %{?_smp_mflags} -C AspectC++ CFLAGS="$BUILDFLAGS" SHARED=1
make %{?_smp_mflags} -C Ag++ CFLAGS="$BUILDFLAGS" SHARED=1

# Generate a config for the binaries to use
export LD_LIBRARY_PATH=$PWD/Puma/lib/linux-release
AspectC++/bin/linux-release/ag++ --gen_config
export PUMA_CONFIG=$PWD/puma.config
ACPP=$PWD/AspectC++/bin/linux-release/ac++
popd

# Phase 2: use the bootstrap tools to do a full build from the unwoven sources
pushd aspectc++

# Weave Puma
make %{?_smp_mflags} -C Puma clean
make %{?_smp_mflags} -C Puma all CFLAGS="$BUILDFLAGS -fPIC" AC=$ACPP
make %{?_smp_mflags} -C Puma doc CFLAGS="$BUILDFLAGS -fPIC" AC=$ACPP
ln -s libPuma.so Puma/lib/linux-release/libPuma.so.0
unset LD_LIBRARY_PATH

# Rebuild ac++ and ag++ with the newly woven Puma
make %{?_smp_mflags} -C AspectC++ CFLAGS="$BUILDFLAGS" SHARED=1
make %{?_smp_mflags} -C Ag++ CFLAGS="$BUILDFLAGS" SHARED=1
popd

%install
# Install the library
make -C aspectc++/Puma install PREFIX=%{buildroot}%{_prefix}
mv %{buildroot}%{_libdir}/libPuma.so %{buildroot}%{_libdir}/libPuma.so.0.0.0
rm -f %{buildroot}%{_libdir}/libPuma.so.0
ln -s libPuma.so.0.0.0 %{buildroot}%{_libdir}/libPuma.so.0
ln -s libPuma.so.0 %{buildroot}%{_libdir}/libPuma.so

# Install the binaries
mkdir -p %{buildroot}%{_bindir}
cp -p aspectc++/AspectC++/bin/linux-release/* %{buildroot}%{_bindir}
cp -p aspectc++/Ag++/bin/linux-release/* %{buildroot}%{_bindir}

# Install the man pages
mkdir -p %{buildroot}%{_mandir}/man1
cp -p %{SOURCE1} %{SOURCE2} %{buildroot}%{_mandir}/man1

# Install missing headers
cp -p aspectc++/Puma/include/Puma/*.inc %{buildroot}%{_includedir}/Puma

# Generate the default config
mkdir -p %{buildroot}%{_libdir}/%{name}
pushd %{buildroot}%{_libdir}/%{name}
LD_LIBRARY_PATH=%{buildroot}%{_libdir} %{buildroot}%{_bindir}/ag++ --gen_config
chmod 0644 puma.config
popd

# Fix permissions
chmod 0755 %{buildroot}%{_libdir}/libPuma.so.0.0.0 %{buildroot}%{_bindir}/*

%check
export LD_LIBRARY_PATH=%{buildroot}%{_libdir}
export PUMA_CONFIG=$PWD/aspectc++.prewoven/puma.config
make -C aspectc++/Puma test
make -C aspectc++/AspectC++ test SHARED=1
make -C aspectc++/Ag++ test SHARED=1

%post -p /sbin/ldconfig

%postun -p /sbin/ldconfig

%triggerin -- gcc-c++
pushd %{_libdir}/%{name}
rm -f puma.config
%{_bindir}/ag++ --gen_config
chmod 0644 puma.config
popd

%files
%doc aspectc++/README.aspectc++ aspectc++/Puma/README
%license aspectc++/Puma/COPYING
%{_bindir}/*
%{_libdir}/libPuma.so.*
%{_mandir}/man1/*
%dir %{_libdir}/%{name}/
%ghost %{_libdir}/%{name}/puma.config

%files devel
%{_includedir}/Puma/
%{_libdir}/libPuma.so

%files doc
%doc aspectc++/Puma/doc/RefMan/html aspectc++/Puma/examples
%doc aspectc++/Puma/doc/UsersManual/UsersManual.pdf
%doc aspectc++/AspectC++/doc/CompilerManual/CompilerManual.pdf
%doc aspectc++/AspectC++/doc/LanguageReference/*.pdf
%doc aspectc++/AspectC++/doc/NoE_ExecModelSurvey/ac++exec-survey.pdf
%doc aspectc++/AspectC++/doc/NoE_ShortSurvey/ac++lang-survey.pdf
%doc aspectc++/AspectC++/doc/QuickRef/ac++quickref.pdf

%changelog
* Tue Oct 21 2014 Jerry James <loganjerry@gmail.com> - 1.2-3
- The -devel subpackage needs to Require lexertl-devel
- Package missing header files
- Use triggers to update the config file instead of rebuilding for every gcc
- Merge aspectc++ into the main package so ag++ is available for the trigger

* Wed Oct 15 2014 Jerry James <loganjerry@gmail.com> - 1.2-2
- Do not fix the FSF's address in the license file (or anywhere else)
- Drop -doc subpackage dependencies
- Add dependencies from main package to libstdc++-devel and glibc-devel
- Move puma.config to the main package, and put it in _libdir/libpuma

* Fri Aug 29 2014 Jerry James <loganjerry@gmail.com> - 1.2-1
- Initial RPM
